#ifndef GENERAL__READ_WRITE_LOCK_HH
#define	GENERAL__READ_WRITE_LOCK_HH

#include "general/config.h"

#include <stdexcept>
#include <string>

#include "general/Thread.hh"

namespace General
{
  class ReadWriteLock
  {
  public:
    enum mode_type
    {
      READ,
      WRITE
    };

    class BusyError
      : public std::runtime_error
    {
    public:
      BusyError( );
    private:
      std::string msg( );
    };

    class TimedoutError
      : public std::runtime_error
    {
    public:
      TimedoutError( );
    private:
      std::string msg( );
    };

    class Sync
    {
    public:
#ifdef HAVE_PTHREAD_RW_LOCK
      typedef pthread_rwlock_t	lock_type;
#else /* HAVE_PTHREAD_RW_LOCK */
      typedef pthread_mutex_t	lock_type;
#endif /* HAVE_PTHREAD_RW_LOCK */
      static lock_type	INITIALIZER;

      Sync( );
      Sync( lock_type& Source );
      Sync( lock_type* Source );

      void Lock( );
      void Unlock( );
      void ReadLock( );
      void ReadUnlock( );

    private:
      lock_type		m_lock;
      lock_type*	m_lock_ptr;

      Sync( const Sync& Source );
      Sync& operator=( const Sync& Source );
    };

    typedef Sync::lock_type	lock_type;

    ReadWriteLock( lock_type& Lock, mode_type Mode,
		   const char* Filename = "",
		   int Linenum = 0 );

    ReadWriteLock( lock_type& Lock, mode_type Mode,
		   bool TryLock,
		   const char* Filename = "",
		   int Linenum = 0 );

    ReadWriteLock( lock_type& Lock, mode_type Mode, int Timeout,
		   const char* Filename = "",
		   int Linenum = 0 );

    virtual ~ReadWriteLock( );

    static void Destroy( lock_type& Lock );
    static void Initialize( lock_type& Lock );

    void Modify( mode_type Mode,
		 const char* Filename = "",
		 int Linenum = 0 );

    void Release( const char* Filename = "",
		  int Linenum = 0 );


    static int Interval( int Value );
    static int Timeout( int Value );

    //-------------------------------------------------------------------
    /// \brief Handler for thead cancelation
    ///
    /// \param[in] VLock
    ///     The VLock is a pointer to a ReadWriteLock object that holds
    ///     a lock.
    //-------------------------------------------------------------------
    static void ThreadCancel( void* VLock );

  private:
    ReadWriteLock( const ReadWriteLock& Source );
    const ReadWriteLock& operator=( const ReadWriteLock& Source );

    static void thread_cleanup( void* VLock );

    lock_type*	m_lock;
    mode_type	m_mode;
    int		m_old_cancel_state;
  }; // class ReadWriteLock

  template< ReadWriteLock::mode_type T, typename V >
  class ReadWriteLockVariable
  {
  public:
    ReadWriteLockVariable( ReadWriteLock::lock_type& Baton,
			   int Timeout,
			   V& Variable,
			   const char* Filename = "",
			   int Linenum = 0 );

    ReadWriteLockVariable( ReadWriteLock::lock_type& Baton,
			   int Timeout,
			   const V& Variable,
			   const char* Filename = "",
			   int Linenum = 0 );

    ReadWriteLockVariable( const ReadWriteLockVariable& Source,
			   const char* Filename = "",
			   int Linenum = 0 );
			     
    const V& Var( ) const;

    V& Var( );

  private:
    ReadWriteLockVariable( );

    ReadWriteLock::lock_type&	m_baton;
    int				m_timeout;
    V&				m_variable;
    mutable ReadWriteLock	m_lock;
    mutable bool		m_is_locked;
    const bool			m_is_read_only;

    ReadWriteLock::lock_type& release( ) const;
  };

  template< ReadWriteLock::mode_type T, typename V >
  inline ReadWriteLockVariable< T, V >::
  ReadWriteLockVariable(  ReadWriteLock::lock_type& Baton,
			  int Timeout,
			  V& Variable,
			  const char* Filename,
			  int Linenum )
    : m_baton( Baton ),
      m_timeout( Timeout ),
      m_variable( Variable ),
      m_lock( Baton, T, m_timeout, Filename, Linenum ),
      m_is_locked( true ),
      m_is_read_only( false )
  {
    
  }
  
  template< ReadWriteLock::mode_type T, typename V >
  inline ReadWriteLockVariable< T, V >::
  ReadWriteLockVariable(  ReadWriteLock::lock_type& Baton,
			  int Timeout,
			  const V& Variable,
			  const char* Filename,
			  int Linenum )
    : m_baton( Baton ),
      m_timeout( Timeout ),
      m_variable( const_cast< V& >( Variable ) ),
      m_lock( Baton, T, m_timeout, Filename, Linenum ),
      m_is_locked( true ),
      m_is_read_only( true )
  {
    
  }
  
  template< ReadWriteLock::mode_type T, typename V >
  inline ReadWriteLockVariable< T, V >::
  ReadWriteLockVariable(  const ReadWriteLockVariable& Source,
			  const char* Filename,
			  int Linenum )
    : m_baton( Source.m_baton ),
      m_timeout( Source.m_timeout ),
      m_variable( Source.m_variable ),
      m_lock( Source.release( ), T, m_timeout,
	      Filename, Linenum ),
      m_is_locked( true ),
      m_is_read_only( Source.m_is_read_only )
  {
    
  }

  template< ReadWriteLock::mode_type T, typename V >
  inline const V& ReadWriteLockVariable< T, V >::
  Var( ) const
  {
    if ( m_is_locked == false )
    {
      throw std::logic_error( "ReadWriteLockVariable: data not protected by lock" );
    }
    return m_variable;
  }

  template< ReadWriteLock::mode_type T, typename V >
  inline V& ReadWriteLockVariable< T, V >::
  Var( )
  {
    if ( T == ReadWriteLock::READ )
    {
      throw std::logic_error( "ReadWriteLockVariable: Cannot access variable for writing while only holding a read lock" );
    }
    if ( m_is_read_only )
    {
      throw std::logic_error( "ReadWriteLockVariable: Cannot access constant variable for writing" );
    }
    if ( m_is_locked == false )
    {
      throw std::logic_error( "ReadWriteLockVariable: data not protected by lock" );
    }
    return m_variable;
  }

  template< ReadWriteLock::mode_type T, typename V >
  ReadWriteLock::lock_type& ReadWriteLockVariable< T, V >::
  release( ) const
  {
    m_lock.Release( );	// Reliquish lock
    m_is_locked = false;
    return m_baton;
  }
} // namespace General

#endif	/* GENERAL__READ_WRITE_LOCK_HH */
