/* -*- mode: c++; c-basic-offset: 2; -*- */

#include "general/config.h"

#include <signal.h>
#include <pthread.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

#include "general/AtExit.hh"
#include "general/unordered_map.hh"
#include "general/ldasexception.hh"
#include "general/Thread.hh"

#if ALT_PTHREAD_STACK_MIN
#undef PTHREAD_STACK_MIN
#define PTHREAD_STACK_MIN ALT_PTHREAD_STACK_MIN
#endif /* ALT_PTHREAD_STACK_MIN */

using General::Thread;

static General::Thread::thread_type NULL_THREAD =
  (General::Thread::thread_type)0;

static void
wakeup( int SignalNumber )
{
  ::signal( SignalNumber, wakeup );
}

class thread_registry
{
public:
  typedef Thread::thread_type	key_type;
  typedef Thread* value_type;

  static void Add( key_type, value_type );

  static value_type Get( key_type );

  static inline key_type Self( )
  {
    return pthread_self( );
  }

  static void Remove( key_type );

private:
  friend class General::Thread::Self;

  typedef General::unordered_map< key_type, value_type > thread_registry_type;

  static thread_registry_type	m_storage;
  static MutexLock::lock_type	m_baton;

  static value_type get( key_type );
};

void thread_registry::
Add( thread_registry::key_type Key, thread_registry::value_type Value )
{
  MutexLock	l( m_baton);

  m_storage[ Key ] = Value;
}

thread_registry::value_type thread_registry::
Get( thread_registry::key_type Key )
{
  MutexLock	l( m_baton);

  return get( Key );
}

void thread_registry::
Remove( thread_registry::key_type Key )
{
  MutexLock	l( m_baton);

  m_storage.erase( Key );
}

thread_registry::value_type thread_registry::
get( thread_registry::key_type Key )
{
  thread_registry_type::const_iterator pos( m_storage.find( Key ) );
  if ( pos == m_storage.end( ) )
  {
    std::ostringstream	msg;
    msg << "There is no managed thread with key: " << (void*)Key;

    throw std::range_error( msg.str( ) );
  }
  return pos->second;
}

thread_registry::thread_registry_type thread_registry::m_storage;
MutexLock::lock_type thread_registry::m_baton = MutexLock::Initialize( );

namespace General
{
  //---------------------------------------------------------------------
  // Thread::Self class
  //---------------------------------------------------------------------

  void Thread::Self::
  CancellationCheck( const std::string& Header,
		    const char* File, const int Line )
  {
    thread_registry::value_type v = NULL;
    try
    {
      v = thread_registry::Get( thread_registry::Self( ) );

    }
    catch( ... )
    {
    }
    if ( v )
    {
      v->CancellationCheck( Header, File, Line );
    }
  }

  void Thread::Self::
  CancellationType( cancel_type Type, int AuxInfo )
  {
    thread_registry::value_type v = NULL;

    try
    {
      v = thread_registry::Get( thread_registry::Self( ) );

    }
    catch( ... )
    {
    }
    if ( v )
    {
      v->CancellationType( Type, AuxInfo );
      v->CancellationEnable( ( Type != CANCEL_IGNORE ) );
    }
  }

  //---------------------------------------------------------------------
  // Thread class
  //---------------------------------------------------------------------
  size_t Thread::m_stack_size_default( PTHREAD_STACK_MIN );

  Thread::
  Thread( )
    : m_parent_tid( pthread_self( ) ),
      m_tid( NULL_THREAD ),
      m_key( MutexLock::Initialize( ) ),
      m_cancel_via_signal( 0 ),
      m_cancel_state( false ),
      m_cancel_thread( false ),
      m_cancellation_type( CANCEL_IGNORE )
  {
  }

  Thread::
  ~Thread( )
  {
    Join( );
  }

  bool Thread::
  operator==( const Thread& Source ) const
  {
    return( Source.m_tid == m_tid );
  }

  void Thread::
  Cancel( )
  {
    if ( m_tid != NULL_THREAD )
    {
      m_cancel_thread = true;
      if ( m_cancel_via_signal > 0 )
      {
	Kill( m_cancel_via_signal );
      }
      else
      {
	pthread_cancel( m_tid );
      }
    }
  }

  //---------------------------------------------------------------------
  /// Deliver any pending cancellation requests to the calling thread.
  /// If cancellation is to be done via exception,
  /// then a cancellation exception is throw.
  /// If not, then a call to the appropriate thread library's
  /// cancellation routine is made.
  ///
  /// \note
  ///     This call must only be made by a child thread.
  //---------------------------------------------------------------------
  void Thread::
  CancellationCheck( const std::string& Header,
		     const char* File, const int Line ) const
  {
    if ( ! pthread_equal( m_tid, pthread_self( ) ) )
    {
      return;
    }
    if ( m_cancel_thread )
    {
      switch( m_cancellation_type )
      {
      case CANCEL_IGNORE:
	break;
      case CANCEL_BY_SIGNAL:
      case CANCEL_EXCEPTION:
	throw cancellation( Header, File, Line );
	break;
      case CANCEL_ASYNCHRONOUS:
      case CANCEL_DEFERRED:
	pthread_testcancel( );
	break;
      }
    }
  }

  //---------------------------------------------------------------------
  /// Establish the cancellability of a thread.
  /// If a Value of true is passed, then the thread will allow
  /// cancellation as determined by the call to CancellationType.
  ///
  /// \note
  ///     This call must only be made by a child thread.
  ///
  /// \see
  ///     CancellationType
  //---------------------------------------------------------------------
  void Thread::
  CancellationEnable( bool Value )
  {
    if ( ! pthread_equal( m_tid, pthread_self( ) ) )
    {
      return;
    }
    int oldstate;

    pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );
    m_cancel_state = Value;
    if ( m_cancel_state )
    {
      switch( m_cancellation_type )
      {
      case CANCEL_IGNORE:
      case CANCEL_EXCEPTION:
	// Nothing to do to setup for this one
	break;
      case CANCEL_BY_SIGNAL:
	if ( m_cancel_via_signal > 0 )
	{
	  //---------------------------------------------------------
	  // Make sure the signal handler is registered
	  //---------------------------------------------------------
	  wakeup( m_cancel_via_signal );
	  //---------------------------------------------------------
	  // Setup to cancel via signal.
	  //---------------------------------------------------------
	  sigset_t	signal_set;
	  sigset_t	old_signal_set;

	  sigemptyset( &signal_set );
	  sigaddset( &signal_set, m_cancel_via_signal );

	  //---------------------------------------------------------
	  // Preserve the old signal mask
	  //---------------------------------------------------------
	  pthread_sigmask( SIG_SETMASK, NULL, &old_signal_set );
	  //---------------------------------------------------------
	  // Set the new signal mask
	  //---------------------------------------------------------
	  pthread_sigmask( SIG_UNBLOCK, &signal_set, NULL );
	}
	break;
      case CANCEL_ASYNCHRONOUS:
      case CANCEL_DEFERRED:
	pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, &oldstate );
	break;
      }
    }
    else
    {
      switch( m_cancellation_type )
      {
      case CANCEL_IGNORE:
      case CANCEL_EXCEPTION:
	// Nothing to do to setup for this one
	break;
      case CANCEL_BY_SIGNAL:
	if ( m_cancel_via_signal > 0 )
	{
	  //-------------------------------------------------------------
	  // Make sure the signal handler is registered
	  //-------------------------------------------------------------
	  wakeup( m_cancel_via_signal );
	  //-------------------------------------------------------------
	  // Setup to cancel via signal.
	  //-------------------------------------------------------------
	  sigset_t	signal_set;
	  sigset_t	old_signal_set;

	  sigemptyset( &signal_set );
	  sigaddset( &signal_set, m_cancel_via_signal );

	  //---------------------------------------------------------
	  // Preserve the old signal mask
	  //---------------------------------------------------------
	  pthread_sigmask( SIG_SETMASK, NULL, &old_signal_set );
	  //---------------------------------------------------------
	  // Set the new signal mask
	  //---------------------------------------------------------
	  pthread_sigmask( SIG_BLOCK, &signal_set, NULL );
	}
	break;
      case CANCEL_ASYNCHRONOUS:
      case CANCEL_DEFERRED:
	pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );
	break;
      }
    }
    
  }

  //---------------------------------------------------------------------
  /// Sets a how a thread can be cancelled.
  /// <ul>
  ///   <li><b>CANCEL_ASYNCHRONOUS</b>
  ///       A cancellation request is immediated delivered to the
  ///       thread.
  ///   <li><b>CANCEL_DEFERRED</b>
  ///       A cancellation request is marked pending for the thread
  ///       and the thread is cancelled when it gets to a system
  ///       cancellation point or when CancellationCheck is called.
  ///   <li><b>CANCEL_EXCEPTION</b>
  ///       A cancellation request is marked pending for the thread
  ///       and the thread throws a cancellation exception upon
  ///       calling CancellationCheck.
  /// </ul>
  ///
  /// \note
  ///     This call must only be made by a chivld thread.
  ///
  /// \see
  ///     CancellationEnable, CancellationCheck
  //---------------------------------------------------------------------
  void Thread::
  CancellationType( cancel_type Type, int AuxInfo )
  {
    if ( ( ! pthread_equal( m_tid, pthread_self( ) ) )
	 || ( ( Type != CANCEL_BY_SIGNAL )
	      && ( m_cancellation_type == Type ) )
	 || ( ( Type == CANCEL_BY_SIGNAL )
	      && ( m_cancellation_type == Type )
	      && ( AuxInfo == m_cancel_via_signal ) ) )
    {
      return;
    }
    int oldtype;
    int oldstate;

    pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );
    m_cancel_via_signal = 0;
    m_cancellation_type = Type;
    switch( m_cancellation_type )
    {
    case CANCEL_ASYNCHRONOUS:
      pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype );
      break;
    case CANCEL_DEFERRED:
      pthread_setcanceltype( PTHREAD_CANCEL_DEFERRED, &oldtype );
      break;
    case CANCEL_BY_SIGNAL:
      m_cancel_via_signal = AuxInfo;
      break;
    case CANCEL_EXCEPTION:
    case CANCEL_IGNORE:
      return;
      break;
    }
    //-------------------------------------------------------------------
    // Restore the state of cancellation
    //-------------------------------------------------------------------
    CancellationEnable( m_cancel_state );
  }

  bool Thread::
  IsParentThread( ) const
  {
    return ( pthread_self( ) == m_parent_tid );
  }

  void Thread::
  Detach( ) const
  {
    if ( m_tid != NULL_THREAD )
    {
      thread_registry::Remove( m_tid );
      // Detach the current thread (makes it non-joinable).
      volatile thread_type tmp = m_tid;

      {
	MutexLock	l( m_key, __FILE__, __LINE__ );

	m_tid = NULL_THREAD;
      }
      // thread_registry( ).erase( m_tid );
      pthread_detach( tmp );
    }
  }

  void Thread::
  Join( )
  {
    if ( m_tid != NULL_THREAD )
    {
      bool release( true );

      int status = pthread_join( m_tid, NULL );
      try
      {
	switch( status )
	{
	case ESRCH:
	  throw range_error( "Thread::Join", __FILE__, __LINE__ );
	  break;
	case EINVAL: 
	  throw invalid_argument( "Thread::Join", __FILE__, __LINE__ );
	  break;
	case EDEADLK: 
	  release = false;
	  throw deadlock( "Thead::Join", __FILE__, __LINE__ );
	  break;
	}
      }
      catch( ... )
      {
	if ( release )
	{
	  // get_thread_registry( ).erase( m_tid );
	  
	  MutexLock	l( m_key, __FILE__, __LINE__ );
	  m_tid = NULL_THREAD;
	}
	throw; // rethrow the exception
      }
      {
	// get_thread_registry( ).erase( m_tid );

	MutexLock	l( m_key, __FILE__, __LINE__ );
	m_tid = NULL_THREAD;
      }
    }
  }

  int Thread::
  Kill( int Signal ) const
  {
    if ( m_tid != NULL_THREAD )
    {
      return pthread_kill( m_tid, Signal );
    }
    // Process could not be found since it is null
    return ESRCH;
  }

  int Thread::
  Spawn( )
  {
    return spawn( );
  }

  int Thread::
  spawn( start_function_type StartFunction )
  {
    int retval = 0;
    if ( m_tid == NULL_THREAD )
    {

      pthread_attr_t attr;
      pthread_attr_init (&attr);
      if ( StackSizeDefault( ) > 0 )
      {
	pthread_attr_setstacksize ( &attr, StackSizeDefault( ) );
      }

      retval = pthread_create( &m_tid, &attr, StartFunction, this );
      pthread_attr_destroy (&attr);

      if ( retval != 0 )
      {
	throw std::runtime_error( "General::Thread::Start: failed to start thread" );
      }
    }
    return retval;
  }

  Thread* Thread::
  start_routine( Thread* ThreadSource )
  {
    if ( General::AtExit::IsExiting( ) == false )
    {
      thread_registry::Add( thread_registry::Self( ), ThreadSource );

      try
      {
	ThreadSource->CancellationType( CANCEL_EXCEPTION );
	ThreadSource->CancellationEnable( true );
	ThreadSource->action( );
      }
      catch( const std::exception& exception )
      {
	if ( ThreadSource->IsCancelled( ) )
	{
	  switch( ThreadSource->CancellationType( ) )
	  {
	  case CANCEL_ASYNCHRONOUS:
	  case CANCEL_DEFERRED:
	    thread_registry::Remove( thread_registry::Self( ) );
	    throw;	// Rethrow the cancelation exception
	    break;
	  default:
	    break;
	  }
	}
      }

      thread_registry::Remove( thread_registry::Self( ) );
      return ThreadSource;
    }
    return (Thread*)NULL;
  }

  Thread::cancellation::
  cancellation( const std::string& Header,
	       const char* File, const int Line )
    : std::runtime_error( format( Header, File, Line ) )
  {
  }

  std::string Thread::cancellation::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": thread cancellation."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }

  Thread::deadlock::
  deadlock( const std::string& Header,
	       const char* File, const int Line )
    : std::logic_error( format( Header, File, Line ) )
  {
  }

  std::string Thread::deadlock::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": attempt to join with self."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }

  Thread::invalid_argument::
  invalid_argument( const std::string& Header,
	       const char* File, const int Line )
    : std::invalid_argument( format( Header, File, Line ) )
  {
  }

  std::string Thread::invalid_argument::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": invalid argument."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }

  Thread::range_error::
  range_error( const std::string& Header,
	       const char* File, const int Line )
    : std::range_error( format( Header, File, Line ) )
  {
  }

  std::string Thread::range_error::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": no thread could be found."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }
}
